unit StepperMotor;

interface
uses forms,ExtCtrls,windows,classes,math,Buttons,sysutils;

type

  TRealTimer = class(TThread) // the TRealTimer is an attempt to get realtime capabilities into windows
   private
    IsWaitableTimerEnabled,
    IsActive,
    fEnabled: boolean;
    fInterval: double;
    fOnTimer: TNotifyEvent;
    hTimer: THandle;
    procedure SetEnabled(const Value: boolean);
    procedure SetInterval(const Value: double);
    function GetTicks: double;
   protected
    procedure Execute;override;
   public
    CpuFreq: double;
    property Enabled: boolean read fEnabled write SetEnabled;
    property Interval: double read fInterval write SetInterval; // seconds
    property OnTimer: TNotifyEvent read fOnTimer write fOnTimer;
    constructor Create;
    destructor Destroy;override;
  end;

  TsmDirection = (smNone,smLeft,smRight);
  TsmStepEvent = procedure(Mindex:integer; direction: TsmDirection)of object;
  TsmCommandChange = procedure(Mindex:integer)of object;

  // program stuff
  psmProgramStep = ^TsmProgramStep;
  TsmProgramStep = record
    Steps: integer;
    Frequency: double;  // Hz
    RampTime: double; // seconds
    Direction: TsmDirection;
  end;
  TsmProgramStepArray = array[0..0]of TsmProgramStep;
  psmProgramSteps = ^TsmProgramStepArray;
  // --

  TsmProgram = class(TObject)
    capacity,count: integer;
    Items: psmProgramSteps;
    IsLooping: boolean;
    function Add:psmProgramStep;
    procedure Clear;
    function Insert(index:integer):psmProgramStep;
    procedure Delete(index:integer);
    constructor Create;
    destructor Destroy;override;
  end;

  TStepperMotor = class(TObject)
  private
    StartRampTime: double;
    LastStepTime: double;
    RotateDir: TsmDirection;
    procedure TimerEvent(Sender: TObject);
  public
    FailedCnt,             // error occurred if > 0;
    Mindex: integer;
    State: byte;        //current state of each Motor (2 bits significant)
    OnStep: TsmStepEvent;  // control program feedback

    Timer: TRealTimer;
    StepProgram: TsmProgram;              // list of commands
    OnCmdChange: TsmCommandChange;        // command row advance
    OnProgramFinished: TsmCommandChange;

    ControlFrequency,
    MaxFrequency: double;
    CurrentCommand,
    CurrentStep: integer;
    IsRotating,IsExecuting: boolean;
    StepLabel: TSpeedButton;
    StepCount: integer;

    ResetDirection: TsmDirection;
    procedure ProgramStep;
    procedure Step(Direction: TsmDirection);
    procedure Rotate(Direction: TsmDirection);
    procedure StartProgram(Step: integer);
    procedure Stop; // stops program and rotate
    constructor Create(MotorIndex: integer); //0..3
    destructor Destroy;override;
  end;

implementation

procedure ReaDTimeStampCounter(var n: int64);
asm
  mov ECX,eax
  db $0f,$31 //RDTSC
  mov [ecx],eax
  mov [ecx+4],edx
end;

function LeftStep(state: integer): integer;
begin
  case state of
  0: result:=2;
  3: result:=1;
  2: result:=3;
  else result:=0;
  end;
end;

function RightStep(state: integer): integer;
begin
  case state of
  0: result:=1;
  1: result:=3;
  3: result:=2;
  else result:=0
  end;
end;

{ TStepperMotor }
constructor TStepperMotor.Create(MotorIndex: integer);
begin
  StepCount := 0;
  LastStepTime:=0;
  Mindex:= MotorIndex;

  Timer:= TRealTimer.Create;
  Timer.Enabled:=false;
  Timer.Interval:=1;
  Timer.OnTimer:= TimerEvent;

  ResetDirection := smRight;
  StepProgram:= TsmProgram.Create;
  IsRotating:= false;
  IsExecuting:= false;
  ControlFrequency:=10;
  MaxFrequency:=1000;
  CurrentCommand:=0;
  CurrentStep:=0;
  FailedCnt:=0;
end;

procedure TStepperMotor.ProgramStep;
var t: double;
begin
  if CurrentCommand>=StepProgram.count then begin
    CurrentCommand:=0;
    CurrentStep:=0;
    if not StepProgram.IsLooping or (StepProgram.count=0) then begin
      IsExecuting := false;
      Timer.Enabled:=false;
      if Assigned(OnProgramFinished) then OnProgramFinished(Mindex);
      exit;
    end;
    if Assigned(OnProgramFinished) then OnProgramFinished(Mindex);
  end;

  if CurrentStep>=StepProgram.Items[CurrentCommand].Steps then begin // advance command
    CurrentStep:=0;
    inc(CurrentCommand);
    ProgramStep;
    if assigned(OnCmdChange)then OnCmdChange(Mindex);
    exit;
  end;

  with StepProgram.Items[CurrentCommand] do begin
    if RampTime>0 then begin
      t:= Timer.GetTicks;
      if CurrentStep=0 then begin
        StartRampTime:=t;
        t:= sqrt(RampTime/Frequency);
      end else begin
        t:= sqrt(RampTime/Frequency);
        t:= t/(CurrentStep+1);
      end;
      if t< 1/Frequency then Timer.Interval := 1/Frequency
      else Timer.Interval := t;
    end else begin
      if Frequency=0 then Timer.Interval:= 0
      else Timer.Interval := 1/Frequency;
    end;
    Step(Direction);
  end;
  inc(CurrentStep);
end;

procedure TStepperMotor.TimerEvent(Sender: TObject);
begin
  if IsExecuting then begin
    ProgramStep;
  end else
  if IsRotating then begin
    Step(RotateDir);
  end;
end;

procedure TStepperMotor.Rotate(Direction: TsmDirection);
begin
  IsRotating := True;
  RotateDir:= Direction;
  if ControlFrequency<=0 then Timer.Interval := 10;
  Timer.Interval := 1/ControlFrequency;
  Timer.Enabled := true;
end;

procedure TStepperMotor.StartProgram(Step: integer);
begin
  IsExecuting := true;
  Timer.Interval:=0;
  Timer.Enabled:=true;
  ProgramStep;
end;

procedure TStepperMotor.Step(Direction: TsmDirection);
var t: double;
begin
  if MaxFrequency>0 then begin
    t:= Timer.GetTicks;
    if t<LastStepTime then begin
      Sleep(round(1000/MaxFrequency));
      inc(FailedCnt);
    end;
    LastStepTime := t + 1/MaxFrequency -0.001;
  end;

  if Direction = smRight then begin
    State:= RightStep(State);
    if Assigned(OnStep) then OnStep(Mindex,Direction);
    inc(StepCount);
    if StepLabel<>nil then StepLabel.Caption := IntToStr(StepCount);
  end else if Direction = smLeft then begin
    State:= LeftStep(State);
    if Assigned(OnStep) then OnStep(Mindex,Direction);
    dec(StepCount);
    if StepLabel<>nil then StepLabel.Caption := IntToStr(StepCount);
  end;
end;

procedure TStepperMotor.Stop;
begin
  CurrentCommand:=0;
  CurrentStep:=0;
  IsExecuting:=false;
  IsRotating:=false;
  Timer.Enabled:=false;
end;

destructor TStepperMotor.Destroy;
begin
  Timer.Free;
  StepProgram.Free;
  inherited;
end;


{ TsmProgram }
function TsmProgram.Add:psmProgramStep;
begin
  if count>=capacity then begin
    inc(capacity,64);
    ReallocMem(Items,capacity*sizeof(TsmProgramStep));
  end;
  result:=@Items[count];
  inc(count);
end;

procedure TsmProgram.Clear;
begin
  count:=0;
  capacity:=0;
  FreeMem(Items);
  Items:=nil;
end;

constructor TsmProgram.Create;
begin
  count:=0;
  capacity:=0;
  Items:=nil;
  IsLooping:= false;
end;

procedure TsmProgram.Delete(index:integer);
var i:integer;
begin
  if index<0 then exit;
  if index<count then begin
    dec(count);
    for i:=index to count-1 do begin
      CopyMemory(@Items[i],@Items[i+1],sizeof(TsmProgramStep));
    end;
  end;
end;

destructor TsmProgram.Destroy;
begin
  FreeMem(Items);
end;

function TsmProgram.Insert(index:integer):psmProgramStep;
var i:integer;
begin
  if (index>=count) or (index<0) then begin
    result:=Add;
    exit;
  end;

  if count>=capacity then begin
    inc(capacity,64);
    ReallocMem(Items,capacity*sizeof(TsmProgramStep));
  end;

  for i:=count-1 downto index do begin
    CopyMemory(@Items[i+1],@Items[i],sizeof(TsmProgramStep));
  end;

  result:=@Items[index];
  inc(count);
end;

{ TRealTimer }
constructor TRealTimer.Create;
var n1,n2: int64; t1: cardinal;
begin
  inherited Create(true);
  hTimer:= CreateWaitableTimer(nil,true,nil); // timer of 100ns accuracy, not available under win95
  fEnabled := false;
  fInterval := 0;
  fOnTimer := nil;
  Priority := tpTimeCritical;

  try
    t1:=GetTickCount;
    ReaDTimeStampCounter(n1);
    Sleep(200);
    t1:=GetTickCount-t1;
    ReaDTimeStampCounter(n2);
    CpuFreq := 1000*(n2-n1)/t1;
  except
    CpuFreq := 0;
  end;
end;

destructor TRealTimer.Destroy;
begin
  Enabled := false; // stop the timer
  Terminate;  // set terminated true
  Resume;  // exit the thread loop
  if hTimer<>0 then CloseHandle(hTimer);
  inherited;
end;

procedure PrintLastError;
var lpMsgBuf: PChar;
begin
  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER or FORMAT_MESSAGE_FROM_SYSTEM,nil,GetLastError,0,@lpMsgBuf,0,nil);
  Application.MessageBox(lpMsgBuf,'error',0);
  LocalFree(cardinal(lpMsgBuf));
end;


procedure TRealTimer.Execute;
var t: int64; dt,d: double;
begin
//  Priority := tpHighest;
  Priority := tpTimeCritical;

  while not Terminated do begin
    IsActive:= true;
    dt:=$7fffffff;
    while fEnabled and (fInterval>0) do begin
      d:= GetTicks - dt;
      if (d>0) then begin
        if (d<fInterval) then d:= fInterval-d
        else d:=0.001;
      end else //}
      d:=fInterval;
      t := round(-10000000 * d);
      if hTimer<>0 then begin
        IsWaitableTimerEnabled:=true;
        if not SetWaitableTimer(hTimer,t,0,nil,nil,false) then begin // set a timer of 100ns accuracy
          PrintLastError;
          d:=d*0.1;
        end;
        WaitForSingleObjectEx(hTimer,round(d*10000),true);
        IsWaitableTimerEnabled:=false;
      end else begin
        Sleep(round(d*1000)); //  <- filthy hack for old win95
      end;

      dt:=GetTicks;
      if fEnabled and (fInterval>0) then begin
        if Assigned(fOnTimer) then fOnTimer(self);
      end;
    end;

    IsActive:= false;
    suspend;
  end;
end;

function TRealTimer.GetTicks: double;
var n: int64;
begin
  if CpuFreq=0 then result := GetTickCount*0.001
  else begin
    ReaDTimeStampCounter(n);
    result:= n/CpuFreq;
  end;
end;

procedure TRealTimer.SetEnabled(const Value: boolean);
var t: int64;
begin
  if fEnabled = Value then exit;
  fEnabled := Value;
  if fEnabled then begin
    if (fInterval>0) and Suspended and not isActive then Resume;
  end else begin
    if IsWaitableTimerEnabled and not Suspended and IsActive then begin
      t:=-1;
      if hTimer<>0 then if SetWaitableTimer(hTimer,t,0,nil,nil,false) then begin // signal the timer immediately
        while not Suspended and IsActive do begin // let it happen
          Application.ProcessMessages;
          Sleep(10);
        end;
      end;
    end;
  end;
end;

procedure TRealTimer.SetInterval(const Value: double);
var t: int64;
begin
  if fInterval = Value then exit;
  fInterval := Value;
  if not fEnabled then exit;

  if fInterval>0 then begin
    if Suspended and not isActive then Resume;
  end else begin
    if IsWaitableTimerEnabled and not Suspended and IsActive then begin
      t:=-1;
      if hTimer<>0 then SetWaitableTimer(hTimer,t,0,nil,nil,false); // signal the timer immediately
      while not Suspended and IsActive do begin // let it happen
        Sleep(10);
      end;
    end;
  end;
end;

end.
